import os

from . import utils, session


class BatchDfileRunner(object):
    """
    Batch-execute a list of pre-created d-files.

    Used to execute BMDS on a remote Windows server, if for example you're
    running a library on a non-Windows OS.
    """

    def __init__(self, inputs):
        self.tempfiles = utils.TempFileList()
        self.inputs = inputs
        self.outputs = []
        self.execute()

    def get_outfile(self, dfile, model_name):
        outfile = dfile.replace('.(d)', '.out')
        oo2 = outfile.replace('.out', '.002')

        # if not exponential, exit early
        if 'exponential' not in model_name.lower():
            return outfile

        # side-effect- cleanup other files created by exponential
        if os.path.exists(outfile):
            self.tempfiles.append(outfile)
        if os.path.exists(oo2):
            self.tempfiles.append(oo2)

        # get exponential model prefix
        prefix = model_name.split('-')[1]
        path, fn = os.path.split(outfile)
        outfile = os.path.join(path, prefix + fn)
        return outfile

    def execute(self):

        for obj in self.inputs:

            # get executable path
            exe = session.BMDS.get_model(obj['bmds_version'], obj['model_name']).get_exe_path()

            # write dfile
            dfile = self.tempfiles.get_tempfile(prefix='bmds-dfile-', suffix='.(d)')
            with open(dfile, 'w') as f:
                f.write(obj['dfile'])

            output = {
                'output_created': False,
                'outfile': None,
            }
            try:
                utils.RunProcess([exe, dfile], timeout=20).call()
                outfile = self.get_outfile(dfile, obj['model_name'])
                oo2 = outfile.replace('.out', '.002')
                if os.path.exists(outfile):
                    self.tempfiles.append(outfile)
                    output['output_created'] = True
                    with open(outfile, 'r') as f:
                        output['outfile'] = f.read()
                if os.path.exists(oo2):
                    self.tempfiles.append(oo2)
            except Exception as e:
                raise e
            finally:
                self.tempfiles.cleanup()

            self.outputs.append(output)

        return self.outputs



"""
If we're not working on Windows, we cannot execute BMDS.

The BMDS source-code can be for other operating systems, but the compiled
binaries can yield different results from the same inputs. Thus, we treat the
Windows binaries as the "standard" and therefore we don't attempt to use
the alternative compilations.

Instead, we monkeypatch the executable steps in running a model and a session.
The dfiles are passed via HTTP to a remote Windows server which will execute
the BMDS and return the result .OUT files in a JSON format.

This requires an additional environment variable, BMDS_HOST, which is the host
path for remote execution: e.g., the string "http://12.13.145.167".
"""

import json
import os
import platform
import requests
import sys

from .session import BMDS
from .models.base import BMDModel

__all__ = []


def _get_payload(models):
    return dict(
        inputs=json.dumps([
            dict(
                bmds_version=model.bmds_version_dir,
                model_name=model.model_name,
                dfile=model.as_dfile(),
            ) for model in models
        ])
    )


if platform.system() != 'Windows':

    host = os.environ.get('BMDS_HOST')
    user = os.environ.get('BMDS_USERNAME')
    pw = os.environ.get('BMDS_PASSWORD')
    _request_session = None
    NO_HOST_WARNING = (
        'Using a non-Windows platform; BMDS cannot run natively in this OS.\n'
        'We can make a call to a remote server to execute.\n'
        'To execute BMDS, please specify the following environment variables:\n'
        '  - BMDS_HOST (e.g. http://bmds-server.com)\n'
        '  - BMDS_USERNAME (e.g. myusername)\n'
        '  - BMDS_PASSWORD (e.g. mysecret)\n'
    )

    def _get_requests_session():
        if host is None or user is None or pw is None:
            raise EnvironmentError(NO_HOST_WARNING)

        global _request_session
        if _request_session is None:
            s = requests.Session()
            s.get('{}/admin/login/'.format(host))
            csrftoken = s.cookies['csrftoken']
            s.post('{}/admin/login/'.format(host), {
                'username': user,
                'password': pw,
                'csrfmiddlewaretoken': csrftoken,
            })
            _request_session = s
        return _request_session

    def _set_outputs(model, result):
        model.output_created = result['output_created']
        if model.output_created:
            model.parse_results(result['outfile'])

    def execute_model(self):
        # execute single model
        session = _get_requests_session()
        url = '{}/dfile/'.format(host)
        payload = _get_payload([self])
        resp = session.post(url, data=payload)

        # parse outputs
        result = resp.json()[0]
        _set_outputs(self, result)

    def execute_session(self):
        # submit data
        session = _get_requests_session()
        url = '{}/dfile/'.format(host)
        payload = _get_payload(self.models)
        resp = session.post(url, data=payload)

        # parse results for each model
        jsoned = resp.json()
        for model, result in zip(self.models, jsoned):
            _set_outputs(model, result)

    # print startup error if host is None
    if host is None or user is None or pw is None:
        sys.stderr.write(NO_HOST_WARNING)

    # monkeypatch
    BMDS.execute = execute_session
    BMDModel.execute = execute_model



import re

from . import constants

EXPONENTIAL = 'E'
CTYPES = [constants.CONTINUOUS, constants.CONTINUOUS_INDIVIDUAL, EXPONENTIAL]


class OutputParser(object):

    # regex for finding numeric values
    re_num = r'[/+/-]?[0-9]+[/.]*[0-9]*[Ee+-]*[0-9]*'

    # line-skips by dataset type
    NUM_LINE_SKIPS_PARAMS = {
        constants.CONTINUOUS: 1,
        constants.CONTINUOUS_INDIVIDUAL: 1,
        constants.DICHOTOMOUS: 1,
        constants.DICHOTOMOUS_CANCER: 1,
        EXPONENTIAL: 4,
    }

    NUM_LINE_SKIPS_FIT = {
        constants.CONTINUOUS: 3,
        constants.CONTINUOUS_INDIVIDUAL: 3,
        constants.DICHOTOMOUS: 2,
        constants.DICHOTOMOUS_CANCER: 2,
    }

    NUM_LINE_SKIPS_AIC = {
        constants.CONTINUOUS: 1,
        constants.CONTINUOUS_INDIVIDUAL: 1,
        EXPONENTIAL: 2,
    }

    def __init__(self, output_text, dtype, model_name):
        assert dtype in constants.DTYPES
        if 'exponential' in model_name.lower():
            dtype = EXPONENTIAL

        self.dtype = dtype
        self.output_text = output_text
        self.model_name = model_name
        self.parse_output_file()

    def parse_output_file(self):
        self.output = {}

        self._set_default_values()
        self._import_single_searches()
        self._import_warnings()
        self._import_line_by_line()
        self._calc_residual_of_interest()
        self._cleanup_nans()

    def _set_default_values(self):
        # set name
        self.output['model_name'] = self.model_name

        # initial setup
        vals = ('Chi2', 'df', 'p_value4', 'residual_of_interest')
        for val in vals:
            self.output[val] = -999

    def _import_line_by_line(self):
        if self.dtype in constants.DICHOTOMOUS_DTYPES:
            fit_tbl = (
                'fit_dose',
                'fit_est_prob',
                'fit_estimated',
                'fit_observed',
                'fit_size',
                'fit_residuals',
            )
        elif self.dtype in CTYPES:
            fit_tbl = (
                'fit_dose',
                'fit_size',
                'fit_observed',
                'fit_estimated',
                'fit_stdev',
                'fit_est_stdev',
                'fit_residuals'
            )

        # set empty tables
        for val in fit_tbl:
            self.output[val] = []

        # begin line-by-line
        outs = self.output_text.splitlines()
        if self.dtype in constants.DICHOTOMOUS_DTYPES:
            self._import_dich_vals()
            for i in range(len(outs)):
                if outs[i] == r'       Variable         Estimate        Std. Err.     Lower Conf. Limit   Upper Conf. Limit':  # noqa
                    self._lbl_parameter(outs, i)
                elif outs[i] == r'     Dose     Est._Prob.    Expected    Observed     Size       Residual':  # noqa
                    self._lbl_fit_cont_dich(outs, i, fit_tbl)

        elif self.dtype in constants.CONTINUOUS_DTYPES:
            for i in range(len(outs)):
                if outs[i] == r'       Variable         Estimate        Std. Err.     Lower Conf. Limit   Upper Conf. Limit':  # noqa
                    self._lbl_parameter(outs, i)
                elif outs[i] == r' Dose       N    Obs Mean     Est Mean   Obs Std Dev  Est Std Dev   Scaled Res.':  # noqa
                    self._lbl_fit_cont_dich(outs, i, fit_tbl)
                elif outs[i] == r'   Test    -2*log(Likelihood Ratio)  Test df        p-value    ':  # noqa
                    self._lbl_pvalue(outs, i)
                elif outs[i] == r"            Model      Log(likelihood)   # Param's      AIC":  # noqa
                    self._lbl_aic_cont_exp(outs, i)

        elif self.dtype == EXPONENTIAL:
            for i in range(len(outs)):
                if outs[i] == r'                     Parameter Estimates':
                    self._lbl_parameter(outs, i)
                elif outs[i] == r'     Dose      N         Obs Mean     Obs Std Dev':  # noqa
                    self._lbl_fit_exp(outs, i, 'observed')
                elif outs[i] == r'      Dose      Est Mean      Est Std     Scaled Residual':  # noqa
                    self._lbl_fit_exp(outs, i, 'estimated')
                elif outs[i] == r'     Test          -2*log(Likelihood Ratio)       D. F.         p-value':  # noqa
                    self._lbl_pvalue(outs, i)
                elif outs[i] == r"                     Model      Log(likelihood)      DF         AIC":  # noqa
                    self._lbl_aic_cont_exp(outs, i)

    def _cleanup_nans(self):
        # standardize possible errors
        fields = ('AIC', 'p_value1', 'p_value2', 'p_value3', 'p_value4')
        for field in fields:
            if field in self.output and self.output[field] in ['NA', 'N/A']:
                self.output[field] = -999

    def _calc_residual_of_interest(self):
        bmd = self.output['BMD']
        if bmd > 0 and len(self.output['fit_dose']) > 0:
            diff = abs(self.output['fit_dose'][0] - bmd)
            r = self.output['fit_residuals'][0]
            for j, val in enumerate(self.output['fit_dose']):
                if abs(val - bmd) < diff:
                    diff = abs(val - bmd)
                    r = self.output['fit_residuals'][j]
            self.output['residual_of_interest'] = r

    def _import_single_searches(self):
        """
        Look for simple one-line regex searches common across dataset types.

        If failed, then return -999.

        AIC is only handled in this method for dichotomous continuous and
        exponential and handled via separate methods.
        """
        searches = {
            # (?<!Setting ) is a special case for preventing
            # "Setting BMD = 100*(maximum dose)" matches
            'BMD': r'(?<!Setting )BMD = +(%s)' % self.re_num,
            'BMDL': r'BMDL = +(%s)' % self.re_num,
            'BMDU': r'BMDU = +(%s)' % self.re_num,
            'CSF': r'Multistage Cancer Slope Factor = +(%s)' % self.re_num,
            'AIC': r'AIC: +(%s)' % (self.re_num),
            'model_version': r'Version: ([\d\.]+);',
            'model_date': r'Date: ([\d/]+)',
        }
        for search in searches:
            m = re.search(searches[search], self.output_text)
            if m:
                try:
                    self.output[search] = float(m.group(1))
                except:
                    self.output[search] = m.group(1)
            else:
                self.output[search] = -999

    def _import_warnings(self):
        """
        Add custom warnings found in output files.

        Warnings in output files are searched for using this method; if a
        warning is found then it will be appended to the warnings list.
        """
        warnings = (
            r'Warning: BMDL computation is at best imprecise for these data',
            r'THE MODEL HAS PROBABLY NOT CONVERGED!!!',
            r'BMR value is not in the range of the mean function',
            r'BMD = 100\*\(maximum dose\)',
            r'BMDL computation failed\.',
        )
        self.output['warnings'] = []
        for warning in warnings:
            m = re.search(warning, self.output_text)
            if m:
                self.output['warnings'].append(m.group())

    def _import_dich_vals(self):
        """
        Import simple dichotomous values.

        Dichotomous values are found on one line, therefore one regex will
        return up to three possible matches for the Chi^2, degrees of freedom,
        and p-value.
        """
        m = re.search(r'Chi\^2 = ({0}|\w+) +d.f. = +({0}|\w+) +P-value = +({0}|\w+)'.format(self.re_num),  # noqa
                      self.output_text)
        cw = {1: "Chi2", 2: 'df', 3: 'p_value4'}
        for val in cw:
            try:
                self.output[cw[val]] = float(m.group(val))
            except:
                self.output[cw[val]] = -999

    def _lbl_parameter(self, outs, i):
        self.output['parameters'] = {}
        cw = {
            1: 'estimate',
            2: 'stdev',
            3: '95_low_limit',
            4: '95_high_limit',
        }
        i += self.NUM_LINE_SKIPS_PARAMS[self.dtype]
        while (len(outs[i].split()) > 0):
            vals = outs[i].split()
            self.output['parameters'][vals[0]] = {}
            for j in range(1, len(vals)):
                try:
                    self.output['parameters'][vals[0]][cw[j]] = float(vals[j])
                except:
                    self.output['parameters'][vals[0]][cw[j]] = vals[j]
            i += 1

    def _lbl_fit_cont_dich(self, outs, i, fit_tbl):
        # Line-by-line: find "Goodness  of  Fit" table
        i += self.NUM_LINE_SKIPS_FIT[self.dtype]
        while len(outs[i]) > 1:
            vals = outs[i].split()
            for j in range(len(vals)):
                try:
                    self.output[fit_tbl[j]].append(float(vals[j]))
                except:
                    self.output[fit_tbl[j]].append(vals[j])
            i += 1

    def _lbl_fit_exp(self, outs, i, table_name):
        # Line-by-line: find "Goodness  of  Fit" table - exponential - observed
        i += 2   # next line and dotted lines

        if table_name == 'observed':
            tbl_names = ('fit_dose', 'fit_size', 'fit_observed', 'fit_stdev')
            rng = range(len(outs[i].split()))
        elif table_name == 'estimated':
            tbl_names = ('fit_dose', 'fit_estimated', 'fit_est_stdev', 'fit_residuals')  # noqa
            rng = range(1, len(outs[i].split()))

        while len(outs[i]) > 0:
            vals = outs[i].split()
            for j in rng:  # skip dose, picked up w/ estimated
                try:
                    self.output[tbl_names[j]].append(float(vals[j]))
                except:
                    self.output[tbl_names[j]].apend(vals[j])
            i += 1

    def _lbl_pvalue(self, outs, i):
        # Line-by-line: find p-values (continuous)
        i += 2   # next line and blank line
        while len(outs[i]) > 0:
            vals = outs[i].split()
            if vals[1] in ['5a', '6a', '7a']:  # fix exponentials
                pvalue = '4'
            else:
                pvalue = vals[1]

            if vals[4] == '<':
                self.output['p_value' + pvalue] = '<0.0001'
            else:
                try:
                    self.output['p_value' + pvalue] = float(vals[4])
                except:
                    if vals[4] == '<.0001':
                        self.output['p_value' + pvalue] = '<0.0001'
                    else:
                        self.output['p_value' + pvalue] = vals[4]

            if pvalue == '4':
                self.output['Chi2'] = float(vals[2])
                self.output['df'] = float(vals[3])
            i += 1

    def _lbl_aic_cont_exp(self, outs, i):
        # Line-by-line: find AIC (continuous)
        i += self.NUM_LINE_SKIPS_AIC[self.dtype]
        field = ('fitted', '2', '3', '4', '5')  # continuous is "fitted"
        while len(outs[i]) > 0:
            vals = outs[i].split()
            if vals[0] in field:
                try:
                    self.output['AIC'] = float(vals[3])
                except:
                    self.output['AIC'] = vals[3]
            i += 1



import logging
from subprocess import Popen
import os
import tempfile
import threading


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class TempFileList(list):
    # Maintains a list of temporary files and cleans up after itself

    def get_tempfile(self, prefix='', suffix='.txt'):
        fd, fn = tempfile.mkstemp(prefix=prefix, suffix=suffix)
        os.close(fd)
        self.append(fn)
        return fn

    def cleanup(self):
        for fn in iter(self):
            try:
                os.remove(fn)
            except OSError:
                pass

    def __del__(self):
        self.cleanup()


class RunProcess(threading.Thread):
    """
    Run process and terminate a process given a specified time interval.

    Adapted from: http://stackoverflow/questions/4158502/

    Two input arguments:
        - cmd, the command to execute in subprocess, as a list
        - timeout, the maximum length of time to execute, in seconds

    Example call:
        - RunProcess(["./foo", "arg1"]).call()
    """

    def __init__(self, cmd, timeout=60):
        threading.Thread.__init__(self)
        self.cmd = cmd
        self.timeout = timeout

    def run(self):
        self.p = Popen(self.cmd)
        self.p.wait()

    def call(self):
        self.start()
        self.join(self.timeout)

        if self.is_alive():
            logging.warning("Process stopped; timeout: %s" %
                            ' '.join(self.cmd))
            self.p.terminate()
            self.join()
